/**
 * @author Marius Orcsik <marius@habarnam.ro>
 */

#include <curl/curl.h>
#include <dbus/dbus.h>
#include <event.h>
#include <time.h>
#include "sstrings.h"
#include "structs.h"
#define STB_DS_IMPLEMENTATION
#include "stb_ds.h"
#include "utils.h"
#include "api.h"
#include "smpris.h"
#include "scrobbler.h"
#include "scrobble.h"
#include "sdbus.h"
#include "sevents.h"
#include "ini.h"
#include "configuration.h"

#define HELP_MESSAGE        "MPRIS scrobbler daemon, version %s\n" \
"Usage:\n  %s\t\tStart daemon\n" \
HELP_OPTIONS \
""


static void print_help(const char *name)
{
    fprintf(stdout, HELP_MESSAGE, get_version(), name);
}

/**
 * TODO list
 *  1. Add support to blacklist players (I don't want to submit videos from vlc, for example)
 */
int main (int argc, char *argv[])
{
    int status = EXIT_FAILURE;
    struct configuration config = {0};

    // TODO(marius): make this asynchronous to be requested when submitting stuff
    struct parsed_arguments arguments = {0};
    parse_command_line(&arguments, daemon_bin, argc, argv);
    if (arguments.has_help) {
        print_help(arguments.name);
        status = EXIT_SUCCESS;
        goto _free_arguments;
    }

    load_configuration(&config, APPLICATION_NAME);
    int count = arrlen(config.credentials);
    if (count == 0) { _warn("main::load_credentials: no credentials were loaded"); }
#if 0
    print_application_config(&config);
    //return EXIT_SUCCESS;
#endif

    struct state state = {0};

    if (!state_init(&state, &config)) {
        _error("main::unable to initialize");
        goto _free_state;
    }

    event_base_dispatch(state.events->base);
    status = EXIT_SUCCESS;

_free_state:
    state_destroy(&state);
    configuration_clean(&config);
_free_arguments:
    arguments_clean(&arguments);

    return status;
}
